------------------------------------------------------------------------------------------------
-- Starcraft 2 Model (M3->3ds) Importer version 0.18
-- by NiNtoxicated (madyavic@gmail.com)
-- Imports M3 models into 3ds max via maxscript
-- Created for 3ds max 2010, but should be backwards compatible
-- Based on reworked code from my World of Warcraft M2 importer
--
-- Big thanks to: 
-- Teal (starcraft.incgamers.com) - PHP M3 parser
-- Volcore (http://volcore.limbicsoft.com/)  - Helped figure out vertex flags
-- Witchsong (http://code.google.com/p/libm3/) - Awesome M3 library and for help on sequence data
-- der_Ton (http://www.doom3world.org/) - Great work on MD5 format which is similar to M3
-- Check out http://code.google.com/p/libm3/ for more indepth detail about the M3 format
-- Head to http://www.sc2mapster.com/ for more mod tools and Starcraft 2 content
--
-- If you use any of this code in your own scripts, please credit
-- where it is due.
--
-- Things to do:
-- 1. Clean up messy code
-- 2. Implement 3dsMax GUI properly
-- 3. Automate MPQ extraction
-- 4. Add good animation support
--
------------------------------------------------------------------------------------------------

-- Version History:
-- 0.18 -- April 25th 2010
-- Added support for new MD34 model format
-- Bones now animate correctly
--
-- 0.14 -- April 10th 2010
-- Aligned all submeshes to their appropriate bones 
-- Skinned vertex positioning now based on all weighted bones (thanks to MD5 documentation)
-- Improved animation processing times (bones animated before vertex skinning)
--
-- 0.13 -- April 1st 2010
-- Mesh now binds to bones properly for most models.
--
-- 0.11 -- March 30th 2010
-- Updated Sequence support
-- Added Sequence UI for navigating sequences
-- Shift to more object orientated coding
--
-- 0.098 -- March 26th 2010
-- Added queryBox error for missing textures before creating any scene objects
-- Updated some of the code regarding sequence data
-- Minor code improvements
--
-- 0.096 -- March 22nd 2010
-- Fixed bump mapping (sets to 100 now)
-- 
-- 0.095 -- March 21st 2010
-- Some basic error catching added
-- Fixed initial bone errors
--
-- 0.09 -- March 20th 2010
-- Added primitive animation support
-- Code housekeeping
-- Updated code, improved processing time
--
-- 0.06 -- March 13th 2010
-- Added rollout menu
--
-- 0.05 -- March 11th 2010
-- Added bone support, however mesh not binding to bones properly yet
--
-- 0.04 -- March 6th 2010
-- Major overhaul of code
-- Reworked tag reading and reference reading, can now read nested structures
-- Vastly improved material reading
--
-- 0.03 -- March 4th 2010
-- Vertice normals added 
-- Added material importing
--
-- 0.02 -- February 29th 2010 - March 2nd 2010
-- Made significant improvements to structures and accessing file data
-- Can now (hopefully) import UV data
-- Added submesh functionality
--
-- 0.01 -- February 26th 2010
-- Initial version of M3 import script
--
------------------------------------------------------------------------------------------------

-- ********************
--  SCRIPT DECLARES
-- ********************
-- globals
global head, cmat, mread, mwrite, aread, bstream, cbones
global scrStart, scrEnd -- For script timeStamps
global tread		= #()
global cmesh		= #()

-- Switches
global flipuv_y 				= true -- Flips uv.y coords
global scrExit					= false -- Indicates premature script exit

-- UI switches
global doCreateBones 		= true -- #warning: can slow import
global doTransformBones 	= true
global doCreateSkin 		= true
global doCreateMesh 		= true
global doCreateMats		= true

-- Misc. variables
global bonesize 	= 0.010
global useFPS 	= 30

--Function declaration
fn ReadTagData tagInd flags = ()
fn M3_AnimUI = ()

-- Temp. write file
--global m2emod = "G:\\3dsmax7\\meshes\\ArcaneGolem.m2"
global efile = (GetDir #export+"\\test.m3") -- Destination m3!

-- *******************
--  M3 STRUCTURES
-- ******************* 
struct M3_SeqData
(
	tag, ablock
)

fn getSize type flags:0 =
(
	local size
	case type of
	(
		#REF: 
		(
			case flags of
			(
				#MD33: size = 0x08
				#MD34: size = 0x0C
			)
		)
		#MD33: size = 0x14
		#MODL:
		(
			case flags of
			(
				23: size = 0x240
			)
		)
		#SEQS: size = 0x58
		#STC_: size = 0x8C
		#STG_: size = 0x10
		#STS_: size = 0x18
		#BONE: size = 0x9C
		#VERT: size = 0x20
		#U16_: size = 0x02
		#U32_: size = 0x04
		#CHAR: size = 0x01
		#DIV_: size = 0x24
		#REGN: size = 0x1C
		#BAT_: size = 0x0E
		#MSEC: size = 0x48
		#MATM: size = 0x08
		#MAT_: size = 0xD4
		#LAYR: size = 0x160
		#IREF: size = 0x40
	)
	
	return size
)

-- ***********
--  HELPERS
-- ***********
-- Liberally stolen from my M2 script
fn echo msg =
(
	format "%\n" (msg) to:listener
)

fn ReadFixedString fixedLen=
(
	local str = ""
	for i = 1 to fixedLen do
	(
		str += bit.intAsChar (ReadByte bstream #unsigned)
	)
	return str
)

fn ReadVec bitStream flag array:0 =
(
	local va = #()
	local vec = [0,0,0]
	
	-- Check if array of vectors or single vector
	if array == 0 then 
	(
		count = 1
	)
	else 
	(
		count = array
		-- Initialize array
		va[count] = 0
	)
	
	for i = 1 to count do
	(
		case flag of
		(
			#v3dshort: (vec.x = ReadShort bitStream #unsigned; vec.y = ReadShort bitStream #unsigned; vec.z = ReadShort bitStream #unsigned;)
			#v3dlong: (vec.x = ReadLong bitStream #unsigned; vec.y = ReadLong bitStream #unsigned; vec.z = ReadLong bitStream #unsigned;)
			#v3dbyte: (vec.x = ReadByte bitStream #unsigned; vec.y = ReadByte bitStream #unsigned; vec.z = ReadByte bitStream #unsigned;)
			#quat: (vec = quat 0 0 0 1; vec.x = ReadFloat bitStream; vec.y = ReadFloat bitStream; vec.z = ReadFloat bitStream; vec.w = ReadFloat bitStream;)
			#v4dshort: (vec = [0,0,0,0]; vec.x = ReadShort bitStream #signed; vec.y = ReadShort bitStream #signed; vec.z = ReadShort bitStream #signed; vec.w = ReadShort bitStream #signed;)
			#v4dbyte: (vec = [0,0,0,0]; vec.x = ReadByte bitStream; vec.y = ReadByte bitStream; vec.z = ReadByte bitStream; vec.w = ReadByte bitStream;)
			#v3d: (vec = [0.0,0.0,0.0]; vec.x = ReadFloat bitStream; vec.y = ReadFloat bitStream; vec.z = ReadFloat bitStream;)
			#matrix: 
			(
				vma = #()
				vma[4] = 0
				for j = 1 to 4 do
				(
					local vm = [0,0,0,0]
					vm.x = ReadFloat bitStream; vm.y = ReadFloat bitStream; vm.z = ReadFloat bitStream; vm.w = ReadFloat bitStream;
					vma[j] = vm
				)
				vec = matrix3 vma[1] vma[2] vma[3] vma[4]
			)
		)
		if (array != 0) then (va[i] = vec) else (va = vec)
	)
	
	return va
)

fn SkipBytes bitStream count =
(
	local unknown
	case count of
	(
		2: unknown = ReadShort bitStream #unsigned
		4: unknown = ReadLong bitStream #unsigned
		default:
		(
			for i = 1 to count do
			(
				unknown = ReadByte bitStream #unsigned
			)
		)
	)
)
	
fn LongToString num=
(
	local str = ""
	for i = 1 to 4 do
	(
		str += bit.intAsChar (bit.and num 0xff)
		-- num = bit.shift num -8
		num /= 256
	)
	str
)

fn ShortToFloat &val = -- Short to Float func
(
	if (val > 0) then (val -= 32767) else (val += 32767)
	val /= 32767.0
)

fn M3_Convert_Time t =
(
	t / (1000 / useFPS)
)

fn M3_GetID chunkid =
(
	local strID = chunkid as string
	local tagID = "BLNK" -- Needs to be 4 chars for modification
	for i = 1 to 4 do
	(
		k = 5 - i
		tagID[i] = strID[k]
	)
	id = toUpper tagID -- Uppercase
	return id
)

fn M3_Reset_Globals =
(
	head = mread = bstream = cmesh = undefined
	cmat = #()
	aread = #()
	tread = #()
	cbones = #()
	gc()
)

-- ************************
-- M3 FUNCTIONS
-- ************************ 
fn M3_Open fname flags =
(
	M3_Reset_Globals()
	
	step = "Accessing Model File"
	
	echo ("Opening "+fname+"...")
	bstream = fopen fname flags
	if bstream==undefined then 
	(
		echo "File not found!"
		throw "File not found"
	)
)

fn M3_Close bitStream =
(
	step = "Close"
	fclose bitStream
)

struct M3_Sphere
(
	vmin, vmax, rad, flags,
	
	fn Read bitStream flags =
	(
		local s = M3_Sphere()
		s.vmin = ReadVec bitStream #v3d
		s.vmax = ReadVec bitStream #v3d
		s.rad = ReadFloat bitStream
		if (head.fileID == #MD34) then s.flags = ReadLong bitStream #unsigned
		
		return s
	)
)

struct M3_Tag -- Tag data
(
	dataID, ofsData, nData, flags,
	
	fn Read hinfo =
	(
		step = "Read Tags"
		
		local bm = ftell(bstream) -- bookmark
		-- Pre-initializing arrays speeds up script
		local tr = #()
		tr[hinfo.nTag] = 0
		if (fseek bstream hinfo.ofsTag #seek_set) then
		(	
			for i=1 to hinfo.nTag  do
			(
				local t 		= M3_Tag()
				local chunkid = (ReadFixedString 4)
				t.dataID 	= (M3_GetID chunkid) as name
				t.ofsData	= ReadLong bstream #unsigned
				t.nData		= ReadLong bstream #unsigned
				t.flags		= ReadLong bstream #unsigned
			
				tr[i] = t
			)
		)
		fseek bstream bm #seek_set -- return to bookmark
	
		return tr
	)
)

struct M3_Ref
(
	entries, refid, data = #(), flags,
	
	fn Read flags:0 =
	(
		local r = M3_Ref()
		r.entries 	= ReadLong bstream #unsigned
		r.refid		= ReadLong bstream #unsigned + 1
		if (head.fileID == #MD34) then r.flags = ReadLong bstream #unsigned
		if (flags != #nodata) then
		(
			r.data 		= ReadTagData r.refid flags
			return r.data
		)
		else return r
	)
)

struct M3_AnimRef -- Animation Reference
(
	flags, animflags, animid,
	
	fn Read =
	(
		ar = M3_AnimRef()
		ar.flags = ReadShort bstream #unsigned
		ar.animflags = ReadShort bstream #unsigned
		ar.animid = ReadLong bstream #unsigned
		
		return ar
	)
)

fn M3_Name name flt ext:undefined =
(
	-- Filter name from path
	local mn = filterString name flt
	mn = mn[mn.count]
	local mext = substring mn (mn.count-3) (4)
	-- Remove extension if specified
	if (ext != undefined) then
	(
		if (matchPattern mext pattern:ext ignoreCase:true) then mn = substring mn 1 (mn.count-4)
	)
	return mn
)

struct M3_Header
(
	fileID, ofsTag, nTag, mref = M3_Ref(),
	
	fn Read &header =
	(
		step = "Read header"
		echo "Reading Model Header..." 
	
		header = M3_Header()
		local fileID = (ReadFixedString 4)
		header.fileID = (M3_GetID fileID) as name
		echo ("Model Type: "+header.fileID as string)
		if (header.fileID != #MD33 and header.fileID != #MD34) then
		(
			echo "Not a valid Starcraft 2 model file!"
			throw "Not a valid Starcraft 2 model file!"
		)
		header.ofsTag 	= ReadLong bstream #unsigned
		header.nTag		= ReadLong bstream #unsigned
		header.mref		= M3_Ref.Read flags:#nodata
	)
)

struct M3_MODL
(
	name,  -- max model name
	versID, 
	SEQS, STC, STS, STG, -- anim info
	bones, 
	vflags, verts, DIV, blist,
    bndSphere,
	matm, mat,
	IREF
)

fn M3_Read_MODL flags =
(
	echo "Reading Model Data..."
	mread				= M3_MODL()
	local rsize			= getSize #Ref flags:head.fileID
	local mname 		= M3_Ref.Read()
	mread.name 		= M3_Name mname "\\" ext:".max"
	echo ("Model Name: " + mread.name)	
	mread.versID 		= ReadLong bstream #unsigned
	-- Animation data
	-- Don't read in unneccessary big chunks of data unless transforming bones
	if (doCreateBones and doTransformBones) then 
	(
		echo "Reading Animations..."
		mread.SEQS = M3_Ref.Read()
		-- Potentially massive array of data, better to store it in a different variable
		aread = M3_Ref.Read()
		mread.STG = M3_Ref.Read()
		
		echo "Animations read\n"
	)
	else
	(
		SkipBytes bstream (rsize * 3)
	)
	
	-- SkipBytes
	case (head.fileID) of
	(
		#MD33: SkipBytes bstream 0x14
		default: SkipBytes bstream 0x1C
	)
	
	-- Bones
	if (doCreateBones) then
	(
		echo "Reading bones..."
		mread.bones		= M3_Ref.Read()
		echo "Bones read\n"
	)
	else
	(
		SkipBytes bstream (rsize)
	)
	
	SkipBytes bstream 4
	
	-- Mesh
	if (doCreateMesh) then
	(
		echo "Reading mesh data..."
		mread.vflags		= ReadLong bstream #unsigned
		mread.verts		= M3_Ref.Read flags:mread.vflags
		mread.DIV			= M3_Ref.Read()
		mread.blist			= M3_Ref.Read flags:#ushort
		echo "Mesh data read\n"
	)
	else
	(
		SkipBytes bstream (rsize * 4)
	)
	
	mread.bndSphere = M3_Sphere.Read bstream head.fileID
	
	case (head.fileID) of
	(
		#MD33: 
		(
			case flags of
			(
				20: SkipBytes bstream 0x5C
				23: SkipBytes bstream 0x64
			)
		)
		default:
		(
			case flags of
			(
				20: SkipBytes bstream 0x1C
				23: SkipBytes bstream 0x84
			)
		)
	)
		
	if (doCreateMats) then
	(
		echo "Reading materials..."
		mread.matm	= M3_Ref.Read()
		mread.mat		= M3_Ref.Read()
		echo "Materials read\n"
	)
	else
	(
		SkipBytes bstream (rsize * 2)
	)
	
	-- SkipBytes
	case (head.fileID) of
	(
		#MD33: SkipBytes bstream 0xA8
		default: SkipBytes bstream 0xFC
	)
	
	if (doCreateSkin) then 
	(
		mread.IREF = M3_Ref.Read()
	)
	else
	(
		SkipBytes bstream (rsize)
	)
)

fn M3_Read_Start =
(
	step = "Read Model Data"
	scrExit = false -- reset script exit flag
	scrStart = TimeStamp() -- start tracking script processing time
	M3_Header.Read &head
	tread = M3_Tag.Read head
	mtag = tread[head.mref.refid]
	-- Position bitstream at modl entry... kind of redundant as we should be there already
	fseek bstream mtag.ofsData #seek_set
	
	M3_Read_MODL mtag.flags
)

struct M3_AnimData
(
	d1 = #(), name, d2, maxframes, flags, d4 = #(), bndSphere,
	cstart, cend, -- Max variables
	
	fn Read tag =
	(
		ad = #()
		ad[tag.nData] = 0
		for i = 1 to tag.nData do
		(
			a = M3_AnimData()
			
			SkipBytes bstream 8
			a.name = M3_Ref.Read()
			SkipBytes bstream 4
			local mframes = ReadLong bstream #unsigned
			a.maxframes = M3_Convert_Time mframes
			SkipBytes bstream 0x1C
			a.bndSphere = M3_Sphere.Read bstream head.fileID
			SkipBytes bstream 0x08
			
			ad[i] = a
		)
		
		echo ("nAnims: " + ad.count as string)
		return ad
	)
)

struct M3_AnimBlock -- Sequence Data
(
	frames, flags, fend, keys,
	
	fn Read tag =
	(
		ac = M3_SeqData()
		ac.tag = tag
		ac.ablock = #()
		ac.ablock[tag.nData] = 0
		
		for i = 1 to tag.nData do
		(
			ab 			= M3_AnimBlock()
			ab.frames 	= M3_Ref.Read flags:#ulong
			ab.flags 		= ReadLong bstream #unsigned
			ab.fend		= ReadLong bstream #unsigned
			ab.keys 		= M3_Ref.Read()
			
			ac.ablock[i] = ab
		)
		
		return ac
	)
)

struct M3_SeqCollection --Sequence Transformations Collection (STC)
(
	name, seqind, animid, animref, animoffs, sdata, cstart, cend,
	
	fn Read tag =
	(
		step = "Read Sequence Transformations"
		
		sc = #()
		sc[tag.nData] = 0
		for i = 1 to tag.nData do
		(
			a = M3_SeqCollection()

			a.name 		= M3_Ref.Read()()
			SkipBytes bstream 4
			a.seqind		= ReadShort bstream #unsigned
			SkipBytes bstream 2
			a.animid		= M3_Ref.Read flags:#ulong
			a.animoffs	= M3_Ref.Read flags:#animind
			SkipBytes bstream 4
			-- sequence data array
			sda = #()
			sda[13] = 0
			for i = 1 to 13 do
			(
				s			= M3_Ref.Read()()
				sda[i]	= s
			)
			a.sdata = sda
			
			sc[i] = a
		)
		
		return sc
	)
)

struct M3_AnimLookup
(
	name, stcind,
	
	fn Read tag =
	(
		step = "Read Animation Lookup"
		
		al = #()
		al[tag.nData] = 0
		for i = 1 to tag.nData do
		(
			local a = M3_AnimLookup()
			a.name = M3_Ref.Read()()
			a.stcind = M3_Ref.Read flags:#ulong
			for k = 1 to a.stcind.count do
			(
				-- Must be positive for max arrays later
				a.stcind[k] += 1
			)
			
			al[i] = a
		)
		
		return al
	)
)

struct M3_Bone
(
	name, parent, pos, transid, rotid, rot, scale,
	
	fn Read tag =
	(
		step = "Read Bones"
		
		br = #()
		br[tag.nData] = 0
		for i = 1 to tag.nData do
		(
			local b = M3_Bone()
			SkipBytes bstream 4
			b.name = M3_Ref.Read()()
			--echo ("Bone_" + i as string + ": " + b.name)
			SkipBytes bstream 4
			b.parent = ReadShort bstream #signed + 1
			SkipBytes bstream 2
			b.transid = M3_AnimRef.Read()
			b.pos = ReadVec bstream #v3d
			SkipBytes bstream 16
			b.rotid = M3_AnimRef.Read()
			b.rot = ReadVec bstream #quat
			SkipBytes bstream 28
			b.scale = ReadVec bstream #v3d
			SkipBytes bstream 36
				
			br[i] = b
		)
		echo ("nBones: "+br.count as string)
		
		return br
	)
)


struct M3_Vertex
(
	pos, bw, bi, normal, uv,
	
	fn Read tag flags =
	(
		step = "Read Verts"
		
		local vbytes, vsize, bsize; --vertice/buffer size, many thanks to Volcore for discovering these flags and their meanings!
		if ((bit.and flags 0x40000) == 0) then
		(
			vbytes = 32
			bsize = 4
		)
		else
		(
			vbytes = 36
			bsize = 8
		)
		
		vsize = tag.nData / vbytes
		local vread = #()
		vread[vsize] = 0
		for i=1 to vsize do
		(
			local v = M3_Vertex()
			v.pos = ReadVec bstream #v3d
			
			-- vertex bone weights
			v.bw = #()
			v.bw[4] = 0
			for j = 1 to 4 do 
			(
				local w = ReadByte bstream #unsigned
				v.bw[j] = w / 255.0
			) 
			
			-- vertex bone index
			v.bi = #()
			v.bi[4] = 0
			for j = 1 to 4 do 
			(
				v.bi[j] = 1 + (ReadByte bstream #unsigned) -- 1 + as maxscript arrays are 1 based
			)

			-- First read in compressed normals
			v.normal = #()
			v.normal[4] = 0
			for j = 1 to 4 do 
			(
				v.normal[j] = ReadByte bstream #unsigned
				v.normal[j] = (2 * v.normal[j] / 255.0) - 1
			)
			
			-- Convert to cartesian
			local vn = [0,0,0] -- need point3 when building mesh
			local w = v.normal[4]
			for j = 1 to 3 do
			(
				vn[j] = v.normal[j] / w
			)
			v.normal = vn
			
			-- Define UV conversion
			uvp = [0,0,0]
			local uvm_sml = 2048.0
			for j=1 to 2 do 
			(
				local uv = ReadShort bstream #signed
				uv = uv / uvm_sml
				--if uv > 1 then echo("Warning! Vert["+i as string+"].uv["+j as string+"]: "+uv as string+" is above 1!")
				if (j==2) and (flipuv_y) then (uv = 1 - uv) -- flip Y
				uvp[j] = uv
			)
			v.uv = uvp
			
			SkipBytes bstream bsize
			
			vread[i] = v
		)
		echo ("nVerts: " + vread.count as string)
		
		return vread
	)
)

struct M3_DIV -- Model divisions
(
	faces, smeshes, bat, msec,
	
	fn Read =
	(
		step = "Read DIV"
		
		div = M3_DIV()
		div.faces = M3_Ref.Read flags:#ushort
		div.smeshes = M3_Ref.Read()
		div.bat = M3_Ref.Read()
		div.msec = M3_Ref.Read()

		echo ("nFaces: " + div.faces.count as string)
		echo ("nSubmeshes: " + div.smeshes.count as string)
		
		return div
	)
)

struct M3_Submesh
(
	indverts, nverts, indfaces, nfaces, indbonelu, nbonelu,
	
	fn Read tag =
	(
		step = "Read Submeshes"
		
		smr = #()
		smr[tag.nData] = 0
		for i = 1 to tag.nData do
		(
			sm = M3_Submesh()
			case (head.fileID) of
			(
				#MD33:
				(
					SkipBytes bstream 4
					sm.indverts = ReadShort bstream #unsigned
					sm.nverts = ReadShort bstream #unsigned
				)
				default:
				(
					SkipBytes bstream 8
					sm.indverts = ReadLong bstream #unsigned
					sm.nverts = ReadLong bstream #unsigned
				)
			)

			sm.indfaces = ReadLong bstream #unsigned
			sm.nfaces = ReadLong bstream #unsigned
			SkipBytes bstream 2
			sm.indbonelu = ReadShort bstream #unsigned
			sm.nbonelu = ReadShort bstream #unsigned
			SkipBytes bstream 6
				
			smr[i] = sm
		)
		
		return smr
	)
)

struct M3_Subref -- Submesh refrences
(
	subid, matid,
	
	fn Read tag =
	(
		step = "Read Submesh Material references"
		
		sr= #()
		sr[tag.nData] = 0
		for i = 1 to tag.nData do
		(
			mat = M3_Subref()
			SkipBytes bstream 4
			mat.subid = ReadShort bstream #unsigned + 1
			SkipBytes bstream 4
			mat.matid = ReadShort bstream #unsigned + 1
			SkipBytes bstream 2
			
			sr[i] = mat
		)
		
		return sr
	)
)

struct M3_MATM
(
	matind, matref,
	
	fn Read tag =
	(
		step = "Read MATM"
		
		mtm = #()
		mtm[tag.nData] = 0
		for i = 1 to tag.nData do
		(
			matm = M3_MATM()
			matm.matind = ReadLong bstream #unsigned + 1
			matm.matref = ReadLong bstream #unsigned + 1
			
			mtm[i] = matm
		)
		
		return mtm
	)
)

struct M3_Layr --more work to be done at a later point
(
	name,
	
	fn Read = 
	(
		l = M3_LAYR()
		
		SkipBytes bstream 4
		ln = M3_Ref.Read() -- .dds file
		if (ln.count > 1) then l.name = M3_Name ln "/" else l.name = ln
		SkipBytes bstream 340
		
		return l
	)
)

struct M3_Mat
(
	name, layrs,

	fn Read_Layrs =
	(
		ml = #()
		-- Static 13 layers in Materials
		ml[13] = 0
		
		for j = 1 to 13 do
		(
			ml[j] = M3_Ref.Read()
		)
		
		return ml
	),
	fn Read tag =
	(
		step = "Read Materials"
		
		mr = #()
		mr[tag.nData] = 0
		for i = 1 to tag.nData do
		(
			m = M3_Mat()
			m.name = M3_Ref.Read()
			--echo ("Mat_" + i as string + ": " + m.name)
			SkipBytes bstream 40
			m.layrs = M3_Mat.Read_Layrs()
			SkipBytes bstream 60
			
			mr[i] = m
		)
		
		echo ("nMaterials: " + mr.count as string)	
		return mr
	),
	fn MapChk tname =
	(
		if (tname.count > 1) then
		(
			mcount = mapPaths.count()
			for i = 1 to mcount do
			(
				local tp = mapPaths.get i + "\\"
				
				tex = tp + tname
				if (doesFileExist(tex)) then
				(
					return tex
				)
				else if (i == mapPaths.count()) then
				(
					return 0
				)
			)
		)
		else
		(
			return 1
		)
	),
	fn Check =
	(
		step = "Check Materials"
		
		-- Material bitmap check
		local matlu = mread.DIV.bat
		local ss = stringstream "", fmt = "%\n"
		local ss_f = false
		for i = 1 to matlu.count do
		(
			-- find material referenced in lookup table
			mat_chk = matlu[i].matID
			mid = mread.matm[mat_chk].matref
			m = mread.mat[mid]
			
			for j = 1 to 13 do
			(
				-- Not sure of a more effective way to do this...
				if (j == 1 or j == 3 or j == 4 or j == 10) then
				(
					-- Check if file exists
					t = M3_Mat.MapChk m.layrs[j].name
					-- If it doesn't, add it to error output for later
					if (t == 0) then 
					(
						format fmt m.layrs[j].name to:ss
						-- Trip error switch
						ss_f = true
					)
				)
			)
		)
		
		if (ss_f) then
		(
			-- Error text
			textErr = "The following textures were not found in your 3ds maps directories:\n"
			textErr += ss
			echo textErr
			textErr += "These textures will not be applied to the model. Continue?"
			local t_scrTime = TimeStamp()
			if (queryBox textErr beep:false != true) then
			(
				-- Flip script exit, disable all further script activity
				echo "ERROR# Textures not found"
				scrExit = true
				doCreateBones = doTransformBones = doCreateSkin = doCreateMesh = doCreateMats = false
			)
			else
			(
				-- Correct for time spent dealing with queryBox
				local t_scrTime = TimeStamp() - t_scrTime
				scrStart += t_scrTime
			)
		)
	)
)

struct M3_Animoffs 
(
	aind, sdind,
	
	fn Read =
	(
		a = M3_Animoffs()
		a.aind	= ReadShort bstream #unsigned + 1
		a.sdind 	= ReadShort bstream #unsigned + 1
		
		return a
	)
)

fn M3_Read_Data tag flags = --Reading U32_, I32_, U8__, etc.
(
	local u = #()
	u[tag.nData] = 0
	for i = 1 to tag.nData do
	(
		local a
		case (flags) of
		(
			#animind: 
			(
				a	= M3_Animoffs.Read()
			)
			#ubyte: a = ReadByte bstream #unsigned
			#ushort: a = ReadShort bstream #unsigned
			#ulong: a = ReadLong bstream #unsigned 
		)
		
		u[i] = a
	)
	
	return u
)

fn ReadTagData tagInd flags =
(
	local bm = ftell(bstream) -- bookmark first
	local d = #()
	th = tread[tagInd]

	if ((fseek bstream th.ofsData #seek_set) and (th.nData > 0)) then
	(
		case (th.dataID) of
        (
			#MODL: d = M3_Read_MODL th
			#SEQS: d = M3_AnimData.Read th
			#STC_: d = M3_SeqCollection.Read th
			#STG_: d = M3_AnimLookup.Read th
			#SD4Q: d = M3_AnimBlock.Read th
			#SD3V: d = M3_AnimBlock.Read th
			#SDR3: d = M3_AnimBlock.Read th
			#U8__ : d = M3_Vertex.Read th flags
			#U16_: d = M3_Read_Data th flags
			#U32_: d = M3_Read_Data th flags
			#I32_: d = M3_Read_Data th flags
			#VEC3: d = ReadVec bstream #v3d array:th.nData
			#QUAT: d = ReadVec bstream #quat array:th.nData
			#DIV_: d = M3_DIV.Read()
			#REGN: d = M3_Submesh.Read th
			#BAT_: d = M3_Subref.Read th
			#MATM: d = M3_MATM.Read th
			#MAT_: d = M3_Mat.Read th
			#LAYR: d = M3_Layr.Read()
			#BONE: d = M3_Bone.Read th
			#CHAR: d = ReadFixedString th.nData
			#IREF: d = ReadVec bstream #matrix array:th.nData
			default: d = M3_Read_Data th #ubyte
		)
	)
	else
		"Read Tag failed!"
	
	fseek bstream bm #seek_set -- return to bookmark
	return d
)

-- ***************************
--  MAX SCENE FUNCTIONS
-- ***************************
-- Scene data funcs
fn Set_Scene_Data obj &ind val  =
(
	setAppData obj ind (val as string) 
	ind+=1
)

fn Get_Scene_Data obj &ind  =
(
	ind+=1 
	return getAppData obj (ind-1)
)

fn Set_Scene_Section obj &ind val =
(
 	Set_Scene_Data obj &ind val.name
	Set_Scene_Data obj &ind val.cstart 
	Set_Scene_Data obj &ind val.cend
)

fn Get_Scene_Section obj &ind  =
(
	ss = M3_AnimData()
	ss.name 	= Get_Scene_Data obj &ind
	ss.cstart 	= Get_Scene_Data obj &ind
	ss.cend 		= Get_Scene_Data obj &ind 
	return ss
)

fn M3_Store_Anims =
(
	step = "Storing Animation Data"
	echo "Storing Animation Data in Scene..."

	-- Initiate Scene Object
	if ($_M3_Anim_Data != undefined) then delete($_M3_Anim_Data)
	-- Create scene object to house animation information
	local ad = point size:0.001 pos:[0,0,0]
	ad.name = "_M3_Anim_Data"
	ad.renderable = off
	ad.isHidden = true
	ad.wirecolor = color 8 8 136
	
	local t = 1
	local astart = 2 -- leave keyframe 0 and 1 as bindposes
	local buffer = 2000 -- buffer between animations
	al = mread.STG
	sq = mread.SEQS
	Set_Scene_Data ad &t aread.count
	for i = 1 to al.count do
	(
		a	= sq[i]
		alu = al[i]
		
		for k = 1 to alu.stcind.count do
		(
			stcid = alu.stcind[k]
			st = aread[stcid]
			st.name = (st.name+"["+(i-1) as string+"]")
			st.cstart = astart
			st.cend	= astart + a.maxframes
			
			astart = st.cend + (M3_Convert_Time buffer)
			
			Set_Scene_Section ad &t st
		)
	)
	frameRate = useFPS
)

fn M3_Bone_Depth b =
(
	-- For organising bones based on depth
	if b.parent == 0 then 0
	else return ( 1 + M3_Bone_Depth mread.bones[b.parent] )
)

fn M3_Get_Bone_Depth bones =
(
	-- sort bones by depth
	bd = #()
	bd[bones.count] = 0
	for i=1 to bones.count do
	(
		bonerec = [(M3_Bone_Depth bones[i]), i]
		bd[i] = bonerec
	)
	fn compfn a b = ( if a.x<b.x then return 1; else if a.x>b.x then return -1; else return 0; )
	qsort bd compfn
	
	return bd
)

fn M3_Bones_Bindpose bones mbones =
(
	local bd = M3_Get_Bone_Depth mbones
	
	for i=1 to bones.count do
	(
		local h = bd[i].y
		b = mbones[h]
		cb = bones[h]
		
		in coordsys parent
		(
			cb.scale = b.scale
			cb.rotation = b.rot
			cb.position = b.pos
		)
	)
)

fn M3_Create_Bones =
(
	step = "Create Bones"
	echo "Creating bones..."
	cbones = #()
	
	mb = mread.bones
	for i = 1 to mb.count do
	(
		b = mb[i]
		--echo ("Creating bone_" + i as string + ": " + b.name)
		
		cb = BoneSys.createBone [0,0,0] [0,0,0] [0,0,0.1]
		cb.name 		= b.name
		cb.width 		= bonesize
		cb.height		= bonesize
		cb.showLinks	= true

		append cbones cb
	)
	max views redraw
	
	echo "Setting up bone hierarchy"
	for i=1 to mb.count do
	(
		b = mb[i]
		if b.parent!=0 then
		(
			-- echo ("Setting parent of " + i as string + " to " + (bones_read[i].par+1) as string)
			cbones[i].parent = cbones[b.parent]
		)
	)
	
	echo "Set up bones bindpose"
	M3_Bones_Bindpose cbones mread.bones

	max views redraw
)

-- from the maxscript reference
fn setNodeWorldRotation theNode theRot = 
( 
	in coordsys (transmatrix theNode.transform.pos) 
 		theNode.rotation = theRot
) 

fn M3_Animate_Bone cb ar bf =
(
	-- get transformation array indices
	ind = ar.animoffs[bf].aind
	sind = ar.animoffs[bf].sdind
	sd = ar.sdata[sind]
	
	--echo ("Animating Bone: "+cb.name)
	case (sd.tag.dataID) of
	(
		#SD3V:
		(
			--echo ("Setting up translation for animation " + ar.name + ", " + ar.cend as string + " frames...")
			-- Translation
			bt = sd.ablock[ind]

			for i = 1 to bt.frames.count do
			(
				t = ar.cstart + (M3_Convert_Time bt.frames[i])
				
				with animate on
				(			
					at time t 
					(
						in coordsys parent cb.pos = bt.keys[i] 
					) 
				)
			)
		)
		#SD4Q:
		(
			--echo ("Setting up rotation for animation " + ar.name + ", " + ar.cend as string + " frames...")
			-- Rotation
			br = sd.ablock[ind]
			
			for i = 1 to br.frames.count do
			(
				t = ar.cstart + (M3_Convert_Time br.frames[i])
				
				with animate on
				(
					at time t 
					(
						local par = cb.parent
						if par != undefined then q = par.rotation * br.keys[i] else q = br.keys[i]
						setNodeWorldRotation cb q
						--in coordsys parent cb.rotation = br.keys[i] 
					) 
				)
			)
		)
		-- end of animation handling
	)
)

fn M3_Setup_Anim cb refid =
(
	-- Gotta clean up animation code at some point...very messy :(
	al = mread.STG
	for i = 1 to al.count do
	(
		-- Find STC's for animation
		alu = al[i]
		for k = 1 to alu.stcind.count do
		(
			sind = alu.stcind[k]
			-- 1. Find refid in animid list
			ar = aread[sind]
			bid = ar.animid
			local bfound = 0
			astart = ar.cstart - 1
			
			for i = 1 to bid.count do
			(
				if (bid[i] == refid) then
				(
					bfound = i
					-- more efficient for breaking for loop, 'break()' lags maxscript
					i = bid.count + 1
				)
			)
			
			-- 2. If refid found -> transform bone with corresponding animation data
			-- if not found -> bone refid references no animation data
			if (bfound != 0) then
			(
				M3_Animate_Bone cb ar bfound
			)
		)
	)
)

fn M3_Animate_Bones =
(
	---------------------- animation
	-- sort bones by depth
	mb = mread.bones
	local bd = M3_Get_Bone_Depth mb
	
	echo "Transforming bones..."
	step = "Bone transforms"
	for k=1 to mb.count do
	(
		local i = bd[k].y
		--echo ("Doing bone " + i as string + " (depth: " + bd[k].x as string + ")" )

		--echo ("Transforming bone " + i as string)
		b = mb[i]
		cb = cbones[i]

		-- translation
		M3_Setup_Anim cb b.transid.animid
		-- rotation
		M3_Setup_Anim cb b.rotid.animid

	)
	echo "Bone animation done\n"
)

fn M3_Assign_Maps map texname =
(
	local t = M3_Mat.MapChk texname
	if (isKindOf t string) then
	(
		bt = BitMapTexture()
		bt.filename = t
		map = bt
	)
	
	return map
)

fn M3_Create_Mats =
(
	step = "Create Materials"
	
	echo "Creating Materials..."
	
	matlu = mread.DIV.bat
	cm = multimaterial numsubs:matlu.count
	
	for i = 1 to matlu.count do
	(
		-- find material referenced in lookup table
		mat_chk = matlu[i].matID
		mid = mread.matm[mat_chk].matref
		m = mread.mat[mid]
		
		local sm = standardMaterial name:(m.name)
		cm[i] = sm
		
		sm.diffuseMap = M3_Assign_Maps sm.diffuseMap m.layrs[1].name
		sm.specularMap = M3_Assign_Maps sm.specularMap m.layrs[3].name
		sm.selfIllumMap = M3_Assign_Maps sm.selfIllumMap m.layrs[4].name
		if (sm.specularMap != undefined) then sm.useSelfIllumColor = true
		sm.bumpMap = M3_Assign_Maps sm.bumpMap m.layrs[10].name
		if (sm.bumpMap != undefined) then sm.bumpMapAmount = 100
		
		-- turn on maps in viewports
		showTextureMap sm on
	)
	
	return cm
)

fn M3_Create_Mesh =
(
	step = "Create Mesh"

	echo "Creating mesh..."

	v = #()
	f = #()
	t = #()
	mtl = #()

	vr = mread.verts 
	fr = mread.DIV.faces

	echo "Gathering vertices..."
	-- build vertices and texcoords
	for i = 1 to vr.count do
	(
		append v vr[i].pos
		append t vr[i].uv
	)

	echo "Gathering submeshes..."
	smr = mread.DIV.smeshes
	matlu = mread.DIV.bat
	for s = 1 to smr.count do
	(
		-- Building the Submeshes
		-- build faces by submesh
		-- (each submesh will be a material id)

		sm = smr[s]
		local up = sm.nfaces
		local find = sm.indfaces
		local vind = sm.indverts
		--echo ("sm_"+s as string+".nfaces: "+up as string +" indfaces:" + sm.indfaces as string)

		if ((mod (up) 3) != 0.0 ) then echo "#ERROR sm.tris not a multiple of 3!"
		--else echo "#INFO sm.tris check passed!"
		
		--echo ("Creating Submesh_" + s as string)
		for i = 1 to (up) by 3 do
		(
			local face = [0,0,0]
			for j = 1 to 3 do
			(
				local k = find + i + j - 1
				case (head.fileID) of
				(
					#MD33: face[j] = fr[k] + 1
					default: face[j] = fr[k] + vind + 1
				)
			)
			
			append f face
			append mtl matlu[s].subid
		)
	)

	echo "Creating mesh object"
	cmesh = mesh vertices:v faces:f materialIDs:mtl tverts:t name:(mread.name)
	
	-- if importing mats, apply to mesh
	if doCreateMats then
	(
		cmat = M3_Create_Mats()
		cmesh.material = cmat
		cmesh.material.showInViewport = true
		setMeditMaterial 1 cmat
	)
	
	update cmesh

	-- set normals
	for i = 1 to vr.count do
	(
		setNormal cmesh i vr[i].normal
	)
	
	-- tvert faces
	if vr.count != 0 do
	(
		step = "Build tv's"
		try
		(
			--Set texcoord faces 
			buildTVFaces cmesh false
			for i = 1 to f.count do
			(
				setTVFace cmesh i f[i]
			)
		)
		catch
		(
			echo (getCurrentException())
			throw ()
		)
	)
	update cmesh
	
	-- Bounding Sphere
	local bsphere = (mread.bndSphere)
	local pos = (bsphere.vmin + bsphere.vmax) / 2
	s = Sphere position:pos radius:bsphere.rad
	-- Hide it for now...
	s.isHidden = true

	step = "Create Mesh Done"
)

fn M3_Apply_Skin =
(
	---------------------- skinning
	step = "Create Skin"
	echo "Applying skin to mesh..."
	
	m = cmesh
	select m
	
	max modify mode
	addModifier m (Skin name:"Skin")
	sk = m.modifiers[#Skin]
	
	modPanel.setCurrentObject sk

	-- 1. add all bones to the skin
	for j=1 to cbones.count do
	(
		skinOps.addBone sk cbones[j] 0
	)
	update m
	max views redraw
	
	-- Greatly speeds up vertex weight assignment processing
	disableSceneRedraw()
	-- 2. for each vertex, setup the bone weights
	for j=1 to mread.verts.count do
	(
		v = mread.verts[j]
		--echo (j as string + ": " + v.bi[1] as string + ", " + v.bi[2] as string + ", " + v.bi[3] as string + ", " + v.bi[4] as string \
		--	+ " --- (" + v.bw[1] as string + "," + v.bw[2] as string + "," + v.bw[3] as string + "," + v.bw[4] as string + ")")
		local lbi = #()
		local lbw = #()
		local finalPos = [0,0,0]
		
		for h = 1 to 4 do
		(
			if v.bw[h]>0 then 
			( 
				k = v.bi[h]
				vweight = v.bw[h]
				subm = mread.DIV.smeshes
				local b
				for s = 1 to subm.count do
				(
					sm = subm[s]
					local minv = sm.indverts
					local maxv = minv + sm.nverts + 1
					if j > minv and j < maxv then
					(
						bind = sm.indbonelu + k
						b = mread.blist[bind] + 1
					)
				)
				
				
				at time 0
				(
					local pos
					cb = cbones[b]
					
					pos = v.pos * mread.IREF[b]
					pos = pos * cb.scale
					pos = pos * (inverse cb.rotation)
				
					finalPos += (cbones[b].pos + pos) * vweight
				)

				append lbi b
				append lbw v.bw[h]
			)
		)
		
		-- move vertex to bindpose position
		meshop.setVert m j finalPos
		
		--echo ("j["+j as string+"].lbi: " + lbi as string + " lbw: " + lbw as string)

		skinOps.ReplaceVertexWeights sk j lbi lbw
	)
	
	enableSceneRedraw()
	
	update m
	redrawViews()
	step = "Skinning complete"
)

fn M3_Zoom_Model =
(
	step = "Zoom Viewports to Model"
	
	viewport.ZoomToBounds true cmesh.min cmesh.max
)
-- ********
--  MAIN
-- ********
-- Import
fn M3I_Main file =
(	
	-- Read Model Data
	echo "\n====== Import Starting ======\n"
	M3_Open file "rb"
	M3_Read_Start()
	M3_Close(bstream)
	if (doCreateMats and doCreateMesh) then M3_Mat.Check()
	
	-- Max Scene Objects
	if (scrExit != true) then echo "Creating Scene objects..."
	-- Bones
	if doCreateBones then
	(
		Try 
		(
			M3_Create_Bones()
		)
		Catch
		(
			echo "ERROR# Bone import failed!"
			echo ("****"+getCurrentException()+"****")
			doCreateBones = false
			doCreateSkin = false
			doTransformBones = false
		)
	)
	-- Bone Animation
	if doCreateBones and doTransformBones then
	(
		Try
		(
			M3_Store_Anims()
			M3_Animate_Bones()
			M3_AnimUI()
		)
		Catch
		(
			echo "ERROR# Bone animation failed!"
			echo ("****"+getCurrentException()+"****")
			doTransformBones = false
		)
	)
	-- Mesh
	if doCreateMesh then
	(
		Try
		(
			M3_Create_Mesh()
			M3_Zoom_Model()
		)
		Catch
		(
			echo "ERROR# Mesh import failed!"
			echo ("****"+getCurrentException()+"****")
			doCreateMesh = false
			doCreateSkin = false
		)
	)
	-- Skin
	if doCreateBones and doCreateSkin and doCreateMesh then
	(
		Try 
		(
			M3_Apply_Skin()
		)
		Catch
		(
			echo "ERROR# Skin creation failed!"
			echo ("****"+getCurrentException()+"****")
			doCreateSkin = false
			-- Fix viewport rendering if skinning failed during vertex weighting
			if isSceneRedrawDisabled() then enableSceneRedraw()
		)
	)
	
	-- Check if script exit has been flipped
	if (scrExit) then
	(
		echo "======= Import Aborted =======\n"
	)
	else
	(
		-- Calculate Script Execution time
		scrEnd = TimeStamp()
		format "Import took % seconds\n" ((scrEnd - scrStart) / 1000.0)
		echo "====== Import Successful ======\n"
	)

	setCommandPanelTaskMode #utility
	max views redraw
	M3_Reset_Globals()
)

-- *********************************
--  USER INTERFACE FUNCTIONS
-- *********************************
-- Global UI Funcs
fn uiOpenFile ftypes &fname =
(
	local ret = getOpenFileName types:ftypes filename:fname
	if ret != undefined then fname = ret
)

utility m3imp "M3 - Import"
(
	group "Model Filename"
	(
		button bOpenFile "Open..."
		edittext tFileName
	)
	
	group "Settings"
	(
		checkBox chkImportMesh "Import Mesh" checked:(doCreateMesh) tooltip:"Imports Model Geometry"
		checkBox chkImportMats "Import Materials" checked:(doCreateMats) tooltip:"Imports Model Materials"
		checkBox chkImportSkin	"Create Skin" checked:(doCreateSkin) enabled:(doCreateBones and doCreateMesh) tooltip:"Deforms mesh through bones"
		checkBox chkImportBones "Import Bones" checked:(doCreateBones) tooltip:"Imports Bones from Model (buggy)"
		checkBox chkTransformBones "Animate Bones" checked:(doTransformBones) enabled:(doCreateBones) tooltip:"Animates model bones"
		spinner spnFPS "FPS:" range:[1,1000,30] enabled:(doTransformBones) type:#integer
	)
	
	button bImport "Import" height:35 width:100
	
	on m3imp open do
	(
		if (doCreateBones != true) then
		(
			chkImportSkin.checked = false
			chkTransformBones.checked = false
		)
		if (chkTransformBones.checked != true) then
		(
			spnFPS.enabled = false
		)
		if (doCreateMesh != true) then
		(
			chkImportSkin.checked = false
		)
	)
	on chkImportMesh changed false do
	(
		chkImportSkin.checked = false
		chkImportSkin.enabled = false
		chkImportMats.checked = false
		chkImportMats.enabled = false
	)
	on chkImportMesh changed true do
	(
		if (chkImportBones.checked != false) then
		(
			chkImportSkin.checked = true
			chkImportSkin.enabled = true
		)
		
		chkImportMats.checked = true
		chkImportMats.enabled = true
	)
	on chkImportBones changed false do
	(
		chkTransformBones.checked = false 	
		chkTransformBones.enabled = false
		chkImportSkin.checked = false
		chkImportSkin.enabled = false
		spnFPS.enabled = false
	)
	on chkImportBones changed true do
	(
		chkTransformBones.enabled	= true
		if (chkImportMesh.checked != false) then
		(
			chkImportSkin.checked = true
			chkImportSkin.enabled = true
		)
		spnFPS.enabled = true
	)
	on chkTransformBones changed true do
	(
		spnFPS.enabled = true
	)
	on chkTransformBones changed false do
	(
		spnFPS.enabled = false
	)
	on bOpenFile pressed do
	(
		uiOpenFile "M3 model (*.m3)|*.m3|All Files|*.*|" &tFileName.text
	)
	on bImport pressed do
	(
		local fname = tFilename.text
		if doesFileExist(fname) then
		(
			doCreateMesh = chkImportMesh.checked
			doCreateSkin = chkImportSkin.checked
			doCreateMats = chkImportMats.checked
			doCreateBones = chkImportBones.checked
			doTransformBones = chkTransformBones.checked
			useFPS = spnFPS.value
			M3I_Main(fname)
			if (scrExit != true) then m3imp.open = false
		)
	)
)

utility animUI "M3 - Sequences"
(
	local animList = #("None")
	local aobj = $_M3_Anim_Data
	local adata = #()
	
	dropdownlist	dlAnim "Animation:" items:#("None")
	edittext		edtStart "Time Start:" enabled:false
	edittext		edtEnd "Time End: " enabled:false
	
	fn Selected sel =
	(
		local ad = adata[sel-1]
		animationRange = interval (ad.cstart as integer) (ad.cend as integer)
		edtStart.text = ad.cstart
		edtEnd.text = ad.cend
	)
	
	fn NoneSelected =
	(
		if (adata.count > 0) then 
		(
			animationRange = interval 0 (adata[adata.count].cend as integer)
			edtEnd.text = adata[adata.count].cend
		)
		else 
		(
			animationRange = interval 0 100
			edtEnd.text = "100"
		)
		edtStart.text = "0"
		edtStart.enabled = false
		edtEnd.enabled = false
	)
	
	fn List_Repop  =
	(
		animList = dlAnim.items
		animList[adata.count+1] = 0
		animList[1] = "None"
		for i = 1 to adata.count do
		(
			--echo (anim_data[i].name)
			animList[i+1] = adata[i].name
		)
		
		rsel = random 1 adata.count
		
		dlAnim.items = animList
		dlAnim.selection = rsel + 1
		Selected dlAnim.selection
	)
	
	fn List_Pop =
	(
		if (aobj != undefined) do
		(	
			--reset vars
			adata = #()
			
			local t = 1
			local animCount = (Get_Scene_Data aobj &t) as integer
			adata[animCount] = 0
			for h = 1 to animCount do
			(
				adata[h] = Get_Scene_Section aobj &t
			)
			
			List_Repop()
		)			
	)
	
	on animUI open do
	(
		List_Pop()
	)
	on dlAnim selected i do
	(
		if (adata.count > 0) then
		(
			if (i == 1) then
			(
				NoneSelected()
			)
			else
			(
				Selected i
			)
		)
	)
)

fn M3_AnimUI =
(
	-- Reset
	if (animUI != undefined) do closeUtility animUI
	openUtility animUI
)